#pragma once

#ifndef __FMOPL_H__
#define __FMOPL_H__

/* select output bits size of output : 8 or 16 */
#define OPL_SAMPLE_BITS 16

#define OPL_TYPE_WAVESEL   0x01  /* waveform select     */

/* ---------- Generic interface section ---------- */
#define OPL_TYPE_YM3812 (OPL_TYPE_WAVESEL)

/* compiler dependence */
typedef unsigned char	UINT8;   /* unsigned  8bit */
typedef unsigned short	UINT16;  /* unsigned 16bit */
typedef unsigned int	UINT32;  /* unsigned 32bit */
typedef signed char		INT8;    /* signed  8bit   */
typedef signed short	INT16;   /* signed 16bit   */
typedef signed int		INT32;   /* signed 32bit   */

#if (OPL_SAMPLE_BITS==16)
typedef INT16 OPLSAMPLE;
#endif
#if (OPL_SAMPLE_BITS==8)
typedef INT8 OPLSAMPLE;
#endif

typedef struct{
	UINT32	ar;			/* attack rate: AR<<2           */
	UINT32	dr;			/* decay rate:  DR<<2           */
	UINT32	rr;			/* release rate:RR<<2           */
	UINT8	KSR;		/* key scale rate               */
	UINT8	ksl;		/* keyscale level               */
	UINT8	ksr;		/* key scale rate: kcode>>KSR   */
	UINT8	mul;		/* multiple: mul_tab[ML]        */

	/* Phase Generator */
	UINT32	Cnt;		/* frequency counter            */
	UINT32	Incr;		/* frequency counter step       */
	UINT8   FB;			/* feedback shift value         */
	INT32   *connect1;	/* slot1 output pointer         */
	INT32   op1_out[2];	/* slot1 output for feedback    */
	UINT8   CON;		/* connection (algorithm) type  */

	/* Envelope Generator */
	UINT8	eg_type;	/* percussive/non-percussive mode */
	UINT8	state;		/* phase type                   */
	UINT32	TL;			/* total level: TL << 2         */
	INT32	TLL;		/* adjusted now TL              */
	INT32	volume;		/* envelope counter             */
	UINT32	sl;			/* sustain level: sl_tab[SL]    */
	UINT8	eg_sh_ar;	/* (attack state)               */
	UINT8	eg_sel_ar;	/* (attack state)               */
	UINT8	eg_sh_dr;	/* (decay state)                */
	UINT8	eg_sel_dr;	/* (decay state)                */
	UINT8	eg_sh_rr;	/* (release state)              */
	UINT8	eg_sel_rr;	/* (release state)              */
	UINT32	key;		/* 0 = KEY OFF, >0 = KEY ON     */

	/* LFO */
	UINT32	AMmask;		/* LFO Amplitude Modulation enable mask */
	UINT8	vib;		/* LFO Phase Modulation enable flag (active high)*/

	/* waveform select */
	UINT16	wavetable;
} OPL_SLOT;

typedef struct{
	OPL_SLOT SLOT[2];
	/* phase generator state */
	UINT32  block_fnum;	/* block+fnum                   */
	UINT32  fc;			/* Freq. Increment base         */
	UINT32  ksl_base;	/* KeyScaleLevel Base step      */
	UINT8   kcode;		/* key code (for key scaling)   */
} OPL_CH;

/* OPL state */
typedef struct fm_opl_f {
	/* FM channel slots */
	OPL_CH	P_CH[9];				/* OPL/OPL2 chips have 9 channels*/

	UINT32	eg_cnt;					/* global envelope generator counter    */
	UINT32	eg_timer;				/* global envelope generator counter works at frequency = chipclock/72 */
	UINT32	eg_timer_add;			/* step of eg_timer                     */
	UINT32	eg_timer_overflow;		/* envelope generator timer overlfows every 1 sample (on real chip) */

	UINT8	rhythm;					/* Rhythm mode                  */

	UINT32	fn_tab[1024];			/* fnumber->increment counter   */

	/* LFO */
	UINT8	lfo_am_depth;
	UINT8	lfo_pm_depth_range;
	UINT32	lfo_am_cnt;
	UINT32	lfo_am_inc;
	UINT32	lfo_pm_cnt;
	UINT32	lfo_pm_inc;

	UINT32	noise_rng;				/* 23 bit noise shift register  */
	UINT32	noise_p;				/* current noise 'phase'        */
	UINT32	noise_f;				/* current noise period         */

	UINT8	wavesel;				/* waveform select enable flag  */

	UINT32	T[2];					/* timer counters               */
	UINT8	st[2];					/* timer enable                 */

	UINT8 type;						/* chip type                    */
	UINT8 address;					/* address register             */
	UINT8 status;					/* status flag                  */
	UINT8 statusmask;				/* status mask                  */
	UINT8 mode;						/* Reg.08 : CSM,notesel,etc.    */

	UINT32 clock;					/* master clock  (Hz)           */
	UINT32 rate;					/* sampling rate (Hz)           */
	double freqbase;				/* frequency base               */
	//attotime TimerBase;			/* Timer base time (==sampling time)*/
} FM_OPL;

FM_OPL *OPLCreate(UINT32 clock, UINT32 rate, int type);
void OPLDestroy(FM_OPL *OPL);
void OPLResetChip(FM_OPL *OPL);
int OPLWrite(FM_OPL *OPL,int a,int v);
void OPLWriteReg(FM_OPL *OPL, int r, int v);
unsigned char OPLRead(FM_OPL *OPL,int a);
int OPLTimerOver(FM_OPL *OPL,int c);
void YM3812UpdateOne(FM_OPL *OPL, OPLSAMPLE *buffer, int length);

#endif /* __FMOPL_H__ */
