#include "include/wl_def.h"
#include "include/32x.h"
#include "include/debug_32x.h"

#include <sys/types.h>


boolean AdLibPresent=true, SoundBlasterPresent=true;

SDMode SoundMode;
SMMode MusicMode;
SDSMode DigiMode;

static word *DigiList;

static boolean  SD_Started;
static int      Volume;
static boolean  SPHack;
static boolean  sqActive;

extern int SoundPtrs[LASTSOUND];
extern int SoundLens[LASTSOUND];
extern int SoundPris[LASTSOUND];

extern void *adlibsfx;

#define NUM_SFX 8

typedef struct {
    byte    *data;
    int     length;
    int     lvol, rvol;
} Sfx;

Sfx __attribute__((aligned(16))) sound_vars[NUM_SFX];

static int CurDigi[NUM_SFX];
static word CurPri[NUM_SFX];
static int leftchannel, rightchannel;


static void SetSoundLoc(fixed gx, fixed gy);
static boolean SD_PlayDirSound(soundnames sound, fixed gx, fixed gy);


void SD_SetVolume(int vol)
{
	Volume = vol;
}

int SD_GetVolume()
{
	return Volume;
}

void SD_Startup()
{
	if (SD_Started)
		return;

    DigiList = PM_GetPage(ChunksInFile - 1);
	InitDigiMap();

	sqActive = false;

	SD_Started = true;
}

void SD_Shutdown()
{
	if (!SD_Started)
		return;

	SD_MusicOff();
	SD_StopSound();

	SD_Started = false;
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_PlaySound() - plays the specified sound on the appropriate hardware
//
///////////////////////////////////////////////////////////////////////////
boolean SD_PlaySound(soundnames sound)
{
	SPHack = false;
	return SD_PlayDirSound(sound, 0, 0);
}

static boolean SD_PlayDirSound(soundnames sound, fixed gx, fixed gy)
{
	if (DigiMode != sds_Off) {
	if (DigiMap[sound] != -1) {
		int i;

		for (i=0; i<NUM_SFX; i++)
			if (!(MARS_SYS_COMM4 & (1<<i)))
				break;
		if (i == NUM_SFX)
			for (i=0; i<NUM_SFX; i++)
				if (SoundPris[sound] >= CurPri[i])
					break;
		if (i == NUM_SFX)
			return false; // no free channels, and priority not high enough to preempt a playing sound

		sound_vars[i].data = PM_GetSoundPage(SwapInt16(DigiList[(DigiMap[sound] * 2) + 0]));
		sound_vars[i].length = SwapInt16(DigiList[(DigiMap[sound] * 2) + 1]);
		if (SPHack) {
			SPHack = false;
			SetSoundLoc(gx, gy);
		    /* fixed-point 1.0 to 0.5 */
			sound_vars[i].lvol = (16 - leftchannel)<<12;
			sound_vars[i].rvol = (16 - rightchannel)<<12;
		} else {
		    /* fixed-point 1.0 */
			sound_vars[i].lvol = (16 - 0)<<12;
			sound_vars[i].rvol = (16 - 0)<<12;
		}
		CurPri[i] = SoundPris[sound];
		CurDigi[i] = sound;
        while (MARS_SYS_COMM6) ; // wait until slave has responded to any earlier requests
        MARS_SYS_COMM6 = (1<<i); // tells slave to reload sfx info
		return true;
	}
	}

	// not a digital sound effect (or digi-sfx off), try playing AdLib sound effect
	if (SoundMode != sdm_Off) {
		int i;

		for (i=0; i<NUM_SFX; i++)
			if (!(MARS_SYS_COMM4 & (1<<i)))
				break;
		if (i == NUM_SFX)
			for (i=0; i<NUM_SFX; i++)
				if (SoundPris[sound] >= CurPri[i])
					break;
		if (i == NUM_SFX)
			return false; // no free channels, and priority not high enough to preempt a playing sound

		sound_vars[i].data = (byte *)(SoundPtrs[sound] + (int)&adlibsfx);
		sound_vars[i].length = SoundLens[sound];
		if (SPHack) {
			SPHack = false;
			SetSoundLoc(gx, gy);
		    /* fixed-point 1.0 to 0.5 */
			sound_vars[i].lvol = (16 - leftchannel)<<12;
			sound_vars[i].rvol = (16 - rightchannel)<<12;
		} else {
		    /* fixed-point 1.0 */
			sound_vars[i].lvol = (16 - 0)<<12;
			sound_vars[i].rvol = (16 - 0)<<12;
		}
		CurPri[i] = SoundPris[sound];
		CurDigi[i] = sound;
        while (MARS_SYS_COMM6) ; // wait until slave has responded to any earlier requests
        MARS_SYS_COMM6 = (1<<i); // tells slave to reload sfx info
		return true;
	}

	return false;
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_SoundPlaying() - returns the sound number of first playing sound,
//		or 0 if nothing playing
//
///////////////////////////////////////////////////////////////////////////
word SD_SoundPlaying()
{
	int i;

	for (i=0; i<NUM_SFX+1; i++)
		if (MARS_SYS_COMM4 & (1<<i))
			return CurDigi[i];

	return 0;
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_StopSound() - if a sound is playing, stops it
//
///////////////////////////////////////////////////////////////////////////
void SD_StopSound()
{
    MARS_SYS_COMM6 = 0xFFFF; /* force slave to stop playing sfx */
    while (MARS_SYS_COMM4) ;
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_WaitSoundDone() - waits until the current sound is done playing
//
///////////////////////////////////////////////////////////////////////////
void SD_WaitSoundDone()
{
	while (SD_SoundPlaying())
		Debug32xDelay(1);
}

/*
==========================
=
= SetSoundLoc - Given the location of an object, munges the values
=	for an approximate distance from the left and right ear, and puts
=	those values into leftchannel and rightchannel.
=
= JAB
=
==========================
*/

#define ATABLEMAX 15
static const byte righttable[ATABLEMAX][ATABLEMAX * 2] = {
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 0, 0, 0, 0, 0, 1, 3, 5, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 6, 4, 0, 0, 0, 0, 0, 2, 4, 6, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 4, 1, 0, 0, 0, 1, 2, 4, 6, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 5, 4, 2, 1, 0, 1, 2, 3, 5, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 5, 4, 3, 2, 2, 3, 3, 5, 6, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 4, 4, 4, 4, 5, 6, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 5, 6, 6, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8}
};
static const byte lefttable[ATABLEMAX][ATABLEMAX * 2] = {
{ 8, 8, 8, 8, 8, 8, 8, 8, 5, 3, 1, 0, 0, 0, 0, 0, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 6, 4, 2, 0, 0, 0, 0, 0, 4, 6, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 6, 4, 2, 1, 0, 0, 0, 1, 4, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 5, 3, 2, 1, 0, 1, 2, 4, 5, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 6, 5, 3, 3, 2, 2, 3, 4, 5, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 6, 5, 4, 4, 4, 4, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 6, 6, 5, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8}
};

static void SetSoundLoc(fixed gx, fixed gy)
{
	fixed xt, yt;
	int x, y;

// translate point to view centered coordinates
//
	gx -= viewx;
	gy -= viewy;

//
// calculate newx
//
	xt = FixedByFrac(gx,viewcos);
	yt = FixedByFrac(gy,viewsin);
	x = (xt - yt) >> TILESHIFT;

//
// calculate newy
//
	xt = FixedByFrac(gx,viewsin);
	yt = FixedByFrac(gy,viewcos);
	y = (yt + xt) >> TILESHIFT;

	if (y >= ATABLEMAX)
		y = ATABLEMAX - 1;
	else if (y <= -ATABLEMAX)
		y = -ATABLEMAX;
	if (x < 0)
		x = -x;
	if (x >= ATABLEMAX)
		x = ATABLEMAX - 1;

	leftchannel  =  lefttable[x][y + ATABLEMAX];
	rightchannel = righttable[x][y + ATABLEMAX];
}

/*
==========================
=
= SetSoundLocGlobal - Sets up globalsoundx & globalsoundy and then calls
=	UpdateSoundLoc() to transform that into relative channel volumes. Those
=	values are then passed to the Sound Manager so that they'll be used for
=	the next sound played (if possible).
=
==========================
*/

void PlaySoundLocGlobal(word s, int id, fixed gx, fixed gy)
{
	SPHack = true;
	SD_PlayDirSound(s, gx, gy);
}

// tell sound driver the current player moved/turned
void UpdateSoundLoc(fixed x, fixed y, int angle)
{
	// SetSoundLoc uses viewx, viewy, viewcos, viewsin in FillBuffer
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_MusicOn() - turns on the sequencer
//
///////////////////////////////////////////////////////////////////////////
void SD_MusicOn()
{
	sqActive = true;
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_MusicOff() - turns off the sequencer and any playing notes
//
///////////////////////////////////////////////////////////////////////////
void SD_MusicOff()
{
	sqActive = false;

    while (MARS_SYS_COMM0) ; // wait until 68000 has responded to any earlier requests
	MARS_SYS_COMM0 = 0x0400; // tells 68000 to stop AdLib music
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_StartMusic() - starts playing the music pointed to
//
///////////////////////////////////////////////////////////////////////////
void SD_StartMusic(int music)
{
	if (MusicMode!=sdm_Off) {
		SD_MusicOff();
		SD_MusicOn();
	}

	if (sqActive) {
	    while (MARS_SYS_COMM0) ; // wait until 68000 has responded to any earlier requests
        MARS_SYS_COMM2 = music;
		MARS_SYS_COMM0 = 0x0301; // tells 68000 to start music score with repeat
	}
}

void SD_SetDigiDevice(SDSMode mode)
{
	if ((mode==sds_Off) || (mode==sds_SoundBlaster)) {
		DigiMode=mode;
	}
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_SetSoundMode() - Sets which sound hardware to use for sound effects
//
///////////////////////////////////////////////////////////////////////////
boolean SD_SetSoundMode(SDMode mode)
{
	if ((mode==sdm_Off) || (mode==sdm_AdLib)) {
		SoundMode=mode;
		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////
//
//	SD_SetMusicMode() - sets the device to use for background music
//
///////////////////////////////////////////////////////////////////////////
boolean SD_SetMusicMode(SMMode mode)
{
	if (mode==smm_Off) {
		MusicMode=mode;
		SD_MusicOff();
		return true;
	}
	if (mode==smm_AdLib) {
		MusicMode=mode;
		SD_MusicOn();
		return true;
	}
	return false;
}
